#include "floppy.h"
#include "fat12.h"
#include "memory.h"
#include "kernel.h"

/*  */
static void *fat12_fat_table_address ;

/*  */
static void fat12_translate_input_file_name_to_full_file_name( char *input_file_name , char *full_file_name ) ;

void fat12_translate_full_file_name_to_short_file_name( char *full_file_name , char *short_file_name , int *file_name_length )
{
  int count = 0 ;

  for( int i = 0 ; i < 8 ; ++i ){
    if( full_file_name[ i ] != ' ' ){
      ++count ;
      if( full_file_name[ i ] >= 'A' && full_file_name[ i ] <= 'Z' ){
        *short_file_name++ = full_file_name[ i ] - 'A' + 'a' ;
      }
      else{
        *short_file_name++ = full_file_name[ i ] ;
      }
    }
  }

  int has_exe_name = 0 ;
  for( int i = 8 ; i < 11 ; ++i ){
    if( full_file_name[ i ] != ' ' ){
      count++ ;
      if( !has_exe_name ){
        has_exe_name = 1 ;
        *short_file_name++ = '.' ;
        count++ ;
      }
      if( full_file_name[ i ] >= 'A' && full_file_name[ i ] <= 'Z' ){
        *short_file_name++ = full_file_name[ i ] - 'A' + 'a' ;
      }
      else{
        *short_file_name++ = full_file_name[ i ] ;
      }
    }
  }
  *short_file_name = 0 ;
  *file_name_length = count ;
}

void fat12_read_information( unsigned int *free_space )
{
  int count = 0 ;
  
  // ͳļĿ
  struct fat12_file_attribute_struct file_attribute ;
  for( int i = 0 ; i < 0xe0 ; ++i ){
    floppy_read_via_address( ( void * )( 0x2600 + i * 32 ) , &file_attribute , 32 ) ; 
    if( file_attribute.full_name[ 0 ] == 0 ){
      break ;
    }
    else if( file_attribute.attribute != 0xf && ( unsigned char )file_attribute.full_name[ 0 ] != 0xe5 ){
      // ļ0f,ʾǳļһ
      // ͷĸe5,ʾǱɾ,ʹõļĿ
      count++ ;
      memory_copy( file_attribute.full_name , kernel_file_list->full_name[ count - 1 ] , 11 ) ;
      kernel_file_list->pos[ count - 1 ] = i ;
      kernel_file_list->size[ count - 1 ] = file_attribute.how_many_byte ;
    }
  }
  kernel_file_list->count = count ;

  // ͳƿÿռ
  // ɨ fat, 00 ,ʾ.
  unsigned int c = 0 ;
  unsigned short date ;
  unsigned char *p = ( unsigned char * )fat12_fat_table_address ;

  for( int i = 2 ; i < 0xb1f + 2 ; ++i ){ // 0xb1f  FAT12  1.44M ռʹõĴغ
    // һڵλ
    unsigned int pos = i * 12 / 8 ;
    unsigned int current_char = *( p + pos ) ;
    unsigned int next_char = *( p + pos + 1 ) ;
    if( i % 2 ){ // ż,ǵǰֽһֽڵĵλ֯      
      date = ( next_char << 4 ) | ( ( current_char & 0xf0 ) >> 4 ) ;
    }
    else{ // ,ǵǰֽڵĸλ,һֽڵ
      date = ( ( next_char & 0xf ) << 8 ) | current_char ;
    }
    if( !date ){
      c++ ;
    }
  }
  *free_space = c ;
}

void fat12_init( void *fat_table_buffer_address )
{
  /*  fat ڴ */
  floppy_read_via_address( ( void * )0x200 , fat_table_buffer_address , 9 * 512 ) ;
  fat12_fat_table_address = fat_table_buffer_address ;
}

void fat12_translate_input_file_name_to_full_file_name( char *input_file_name , char *full_file_name )
{
  for( int i = 0 ; i < 8 ; ++i ){
    if( *input_file_name && *input_file_name != '.' ){
      if( *input_file_name <= 'z' && *input_file_name >= 'a' ){
        full_file_name[ i ] = *input_file_name - 'a' + 'A' ; // תɴдĸ
      }
      else{
        full_file_name[ i ] = *input_file_name ;
      }
      ++input_file_name ;
    }
    else{
      full_file_name[ i ] = ' ' ;
    }
  }
  if( *input_file_name == '.' ){
    ++input_file_name ;
  }
  // ٻչ,3ַ
  for( int i = 8 ; i < 11 ; ++i ){
    if( *input_file_name ){
      if( *input_file_name <= 'z' && *input_file_name >= 'a' ){
        full_file_name[ i ] = *input_file_name - 'a' + 'A' ; // תɴдĸ
      }
      else{
        full_file_name[ i ] = *input_file_name ;
      }
      ++input_file_name ;
    }
    else{
      full_file_name[ i ] = ' ' ;
    }
  }
}

void fat12_load_file( char *file_name , void *object_address )
{
  /* Ȼʵȫļ */
  char full_file_name[ 11 ] ;
  fat12_translate_input_file_name_to_full_file_name( file_name , full_file_name ) ;

  // ļĿ
  struct fat12_file_attribute_struct file_attribute ;
  for( int i = 0 ; i < 0xe0 ; ++i ){
    floppy_read_via_address( ( void * )( 0x2600 + i * 32 ) , &file_attribute , 32 ) ; //0x91409
    if( file_attribute.full_name[ 0 ] == 0 ){
      // ͷ0,ʾû
      break ;
    }
    else if( memory_is_equal( file_attribute.full_name , full_file_name , 11 ) ){
      break ;
    }
  }
  if( file_attribute.full_name[ 0 ] == 0 ){
    // δҵļ,ֱӷ
    return ;
  }
  // ҵļ,ڿʼȡ
  // غ
  unsigned short cluster = file_attribute.start_cluster_lo16 ;
  unsigned int size = file_attribute.how_many_byte ;

  while( cluster != 0xfff ){ //0x91454
    void *address_in_floppy = ( void * )( 0x4200 + ( cluster - 2 ) * 512 ) ;
    if( size > 512 ){
      floppy_read_via_address( address_in_floppy , object_address , 512 ) ;
      size -= 512 ;
      object_address += 512 ;
    }
    else{
      floppy_read_via_address( address_in_floppy , object_address , size ) ;
      object_address += size ;
    }
    // һڵλ
    unsigned char *p = ( unsigned char * )fat12_fat_table_address ;
    unsigned int pos = cluster * 12 / 8 ;
    unsigned int current_char = *( p + pos ) ;
    unsigned int next_char = *( p + pos + 1 ) ;
    if( cluster % 2 ){ // ż,ǵǰֽһֽڵĵλ֯      
      cluster = ( next_char << 4 ) | ( ( current_char & 0xf0 ) >> 4 ) ;
    }
    else{ // ,ǵǰֽڵĸλ,һֽڵ
      cluster = ( ( next_char & 0xf ) << 8 ) | current_char ;
    }
  }
}
