#include "floppy.h"
#include "io.h"
#include "interrupt.h"
#include "vesa.h"
#include "dma.h"
#include "memory.h"

/******************************************
*                                     *
******************************************/

/******************************************
*                                 *
******************************************/

/* 屣滺ַı */
static void *floppy_buffer_address ;

/* ״̬ */
//static volatile unsigned int floppy_called = 0 ;
static int floppy_called = 0 ;

/******************************************
*                                 *
******************************************/
void floppy_asm_handle_for_floppy_interrupt() ;
void floppy_handle_for_floppy_interrupt() ;

static int floppy_send_byte_return_is_ok( unsigned char byte ) ;
static int floppy_read_byte_return_is_ok( unsigned char *save_read_byte ) ;
static void floppy_recalibrate() ;
static void floppy_translate_address_to_sector( void *address_in_floppy , unsigned int *head , unsigned int *cylinder , unsigned int *sector , unsigned int *offset ) ;
static void floppy_reset() ;
static void floppy_set_specification() ;
static void floppy_start_motor() ;
static void floppy_seek_cylinder( unsigned char head , unsigned char cylinder ) ;
static void floppy_stop_motor() ;
static void floppy_wait_for_interrupt() ;

/******************************************
* Ǻʵ                          *
******************************************/

/* жϺ */
void floppy_wait_for_interrupt()
{
  while( !floppy_called ) ;
  floppy_called = 0 ;
}

/* ȡݺ,ֵǷȡɹ */
int floppy_read_byte_return_is_ok( unsigned char *save_read_byte )
{
  // ﱾӦһ,ԵһʱΪ,false,ʹ,
  // Ĭϲ,ʲȡ޳ȴԼ,ֱΪֹ,ؽ
  while( 1 ){
    unsigned char msr = io_read_from_io_port( 0x3f4 ) ;
    if( ( msr & 0xc0 ) == 0xc0 ){ //  msr ͷλǲ 11
      *save_read_byte = io_read_from_io_port( 0x3f5 ) ;
      return 1 ;
    }
  }
}

/* ݺ,ֵǷͳɹ */
int floppy_send_byte_return_is_ok( unsigned char byte )
{
  // ﱾӦһ,ԵһʱΪ,false,ʹ,
  // Ĭϲ,ʲȡ޳ȴԼ,ֱΪֹ,ؽ
  while( 1 ){
    unsigned char msr = io_read_from_io_port( 0x3f4 ) ;
    if( ( msr & 0xc0 ) == 0x80 ){ //  msr ͷλǲ 10
      io_write_to_io_port( 0x3f5 , byte ) ;
      return 1 ; // ɹ
    }
  }
}

/* ַܹ */
void floppy_read_via_address( void *address_in_floppy , void *address_in_memory , unsigned int how_many_byte )
{
  unsigned int head , cylinder , sector , offset ;
  floppy_translate_address_to_sector( address_in_floppy , &head , &cylinder , &sector , &offset ) ;
  while( how_many_byte > 0 ){  
    // 0x90dc1
    floppy_read( sector , cylinder , head ) ; 
    // ѡõ
    if( how_many_byte > 512 - offset ){
      memory_copy( floppy_buffer_address + offset , address_in_memory , 512 - offset ) ;
      how_many_byte -= 512 - offset ;
      address_in_memory += 512 - offset ;
    }
    else{
      memory_copy( floppy_buffer_address + offset , address_in_memory , how_many_byte ) ;
      how_many_byte = 0 ;
    }
    offset = 0 ;
    sector++ ;
    if( sector == 19 ){
      sector = 0 ;
      head++ ;
      if( head == 2 ){
        head = 0 ;
        cylinder++ ;
      }
    }
  }
}

void floppy_translate_address_to_sector( void *address_in_floppy , unsigned int *head , unsigned int *cylinder , unsigned int *sector , unsigned int *offset )
{
  *sector = ( unsigned int )address_in_floppy / 512 % 18 + 1 ;
  *head = ( unsigned int )address_in_floppy / 512 / 18 % 2 ;
  *cylinder = ( unsigned int )address_in_floppy / 512 / 36 ;
  *offset = ( unsigned int )address_in_floppy % 512 ;
}

/* 趨 */
void floppy_set_specification()
{
  floppy_send_byte_return_is_ok( 3 ) ;
  floppy_send_byte_return_is_ok( 0xdf ) ;
  floppy_send_byte_return_is_ok( 2 ) ;
}

void floppy_reset()
{
  /* λ */
  io_write_to_io_port( 0x3f2 , 0 ) ;
  io_write_to_io_port( 0x3f7 , 0 ) ;
  io_write_to_io_port( 0x3f2 , 0x0c ) ;

  /* ȴж */
  floppy_wait_for_interrupt() ;

  /* ȡؽ*/
  for( int i = 0 ; i < 4 ; ++i ){
    floppy_send_byte_return_is_ok( 8 ) ;
    unsigned char st0 ;
    floppy_read_byte_return_is_ok( &st0 ) ;
    unsigned char cylinder ;
    floppy_read_byte_return_is_ok( &cylinder ) ;
  }
  
  /* 趨 */
  floppy_set_specification() ;
}

/* жϴ */
void floppy_handle_for_floppy_interrupt()
{
  floppy_called = 1 ;
}

/* ʼ */
void floppy_init( void *buffer_address )
{
  /* ж */
  interrupt_close_interrupt() ;

  floppy_called = 0 ;

  /* װж */
  interrupt_install_handle_for_interrupt( 0x26 , floppy_asm_handle_for_floppy_interrupt ) ;

  /* жλ */
  interrupt_set_interrupt_mask_word( 6 , 1 ) ;

  /* ж */
  interrupt_open_interrupt() ;

  /* 趨ַ */
  floppy_buffer_address = buffer_address ;

  /* ʼ */
  floppy_reset() ;
}

/* У׼ͷ */
void floppy_recalibrate()
{
  floppy_send_byte_return_is_ok( 7 ) ; // У׼
  floppy_send_byte_return_is_ok( 0 ) ; // Ͳ
  
  floppy_wait_for_interrupt() ;

  floppy_send_byte_return_is_ok( 8 ) ; // ͻȡж
  unsigned char st0 ;
  floppy_read_byte_return_is_ok( &st0 ) ; // ȡST0
  unsigned char current_cylinder ;
  floppy_read_byte_return_is_ok( &current_cylinder ) ; // ȡPCN
}

void floppy_start_motor()
{
  /*  */
  io_write_to_io_port( 0x3f2 , 0x1c ) ;
}

void floppy_seek_cylinder( unsigned char head , unsigned char cylinder )
{
  floppy_send_byte_return_is_ok( 15 ) ;
  floppy_send_byte_return_is_ok( head << 2 ) ;
  floppy_send_byte_return_is_ok( cylinder ) ;
  
  floppy_wait_for_interrupt() ;

  floppy_send_byte_return_is_ok( 8 ) ;
  unsigned char st0 ;
  floppy_read_byte_return_is_ok( &st0 ) ;
  unsigned char current_cylinder ;
  floppy_read_byte_return_is_ok( &current_cylinder ) ;
}

void floppy_stop_motor()
{
  /* ֹͣ */
  io_write_to_io_port( 0x3f2 , 0 ) ;
}

/* ȡһζһ */
void floppy_read( unsigned char sector , unsigned char cylinder , unsigned char head )
{
  //жϱ־ڵȴжʱʹ
  floppy_called = 0 ;

  //
  floppy_start_motor() ;
  
  //У׼ͷʹ䶨λȷλ
  floppy_recalibrate() ;
  
  //dmaͨ
  dma_open_dma_channel( 2 , floppy_buffer_address , 512 , DMA_READ_MODEL ) ;
  
  //Ͷȡ
  floopy_send_read_commmand(head, cylinder, sector);
  
  //ȴжϲ
  floppy_wait_for_interrupt() ;

  //ȡִеĽ
  floppy_read_result_of_read_command();

  //ֹͣ
  floppy_stop_motor() ;
}

void floppy_send_read_command(unsigned char head, unsigned char cylinder, unsigned char sector)
{
  floppy_send_byte_return_is_ok( 0x66 ) ;
  floppy_send_byte_return_is_ok( head << 2 ) ;
  floppy_send_byte_return_is_ok( cylinder ) ;
  floppy_send_byte_return_is_ok( head ) ;
  floppy_send_byte_return_is_ok( sector ) ;
  floppy_send_byte_return_is_ok( 2 ) ;
  floppy_send_byte_return_is_ok( 18 ) ;
  floppy_send_byte_return_is_ok( 27 ) ;
  floppy_send_byte_return_is_ok( 255 ) ;
}

void floppy_read_result_of_read_command()
{
  for( int i = 0 ; i < 7 ; ++i ){
    unsigned char byte ;
    floppy_read_byte_return_is_ok( &byte ) ;
  }
}