#include "interrupt.h"
#include "vesa.h"
#include "io.h"
#include "message.h"
#include "kernel.h"
#include "mouse.h"

/*****************************************************
*                                          *
*****************************************************/

void mouse_asm_handle_for_mouse_interrupt() ;

void mouse_handle_for_mouse_interrupt() ;

void mouse_enable_mouse() ;

static void *mouse_get_mouse_picture_addr( enum mouse_type_enum mouse_type ) ;

/*****************************************************
* Ƕ                                         *
*****************************************************/

static const int MOUSE_MOUSE_INTERRUPT_NUMBER = 0x2c ;
static const int MOUSE_MOUSE_INTERRUPT_IRQ_NUMBER = 12 ;

static const unsigned short mouse_old_picture[ 32 * 32 ] ;

int ttmp = 0 ;
/*****************************************************
* Ǻʵ                                   *
*****************************************************/

void *mouse_get_mouse_picture_addr( enum mouse_type_enum mouse_type )
{
  switch( mouse_type ){
    case MOUSE_NORMAL_MOUSE :
      return ( void * )0x130000 ;
    case MOUSE_OVER_MOUSE :
      return ( void * )0x131000 ;
    case MOUSE_DOWN_MOUSE :
      return ( void * )0x132000 ;
  }
}

// mouse ʼ
void mouse_init()
{
  //  mouse
  mouse_enable_mouse() ;

  // װжϴ
  interrupt_install_handle_for_interrupt( MOUSE_MOUSE_INTERRUPT_NUMBER , mouse_asm_handle_for_mouse_interrupt ) ;

  // жĻ
  interrupt_set_interrupt_mask_word( MOUSE_MOUSE_INTERRUPT_IRQ_NUMBER , 1 ) ;
}

// mouse жϴ
void mouse_handle_for_mouse_interrupt()
{
  static int x_position = 444 ;
  static int y_position = 300 ;
  static int count = 0 ;
  static int x_sign = 0 ;
  static int y_sign = 0 ;

  static struct message_message_struct message ;
    
  char ch = io_read_from_io_port( 0x60 ) ;

  switch( ++count ){
    case 1 :
      // յǵһֽ
      // ⰴϢ
      message.dose_left_button_down = ch & 0x1 ;
      message.dose_right_button_down = ch & 0x2 ;
	  x_sign = ch & 0x10 ? 0xffffff00 : 0 ;
	  y_sign = ch & 0x20 ? 0xffffff00 : 0 ;
      break ;
    case 2 :
      // յǵڶֽڣ x λ      
      x_position += ( x_sign | ch ) ;
      break ;
    case 3 :
      // յǵֽڣ y λ      
      y_position += -( y_sign | ch ) ;

      // Ϣ͸ں 
      message.message_type = MESSAGE_MOUSE_MESSAGE ;
      message.x_position = x_position ;
      message.y_position = y_position ;
      message_put_message( &kernel_message_queue , message ) ;
      count = 0 ;
      break ;
  }
}

//  mouse 
void mouse_enable_mouse()
{
  //  8042 ̿оƬб  
  //  mouse ӿ 
  io_write_to_io_port( 0x64 , 0xa8 ) ;
  // ֪ͨ 8042 ¸ֽڵķ 0x60 ݽ mouse
  io_write_to_io_port( 0x64 , 0xd4 ) ;
  //  mouse 
  io_write_to_io_port( 0x60 , 0xf4 ) ;
  // ֪ͨ 8042,¸ֽڵķ 0x60 Ӧ 8042 Ĵ
  io_write_to_io_port( 0x64 , 0x60 ) ;
  // ɼ̼ mouse ӿڼж
  io_write_to_io_port( 0x60 , 0x47 ) ;
}

/* ʾ mouse */
void mouse_show_mouse( int x , int y , enum mouse_type_enum mouse_type )
{
  // ȱ mouse ʾǸͼ 
  mouse_save_picture( x , y ) ;
  
  // ٻ mouse
  void *mouse_picture_addr = mouse_get_mouse_picture_addr( mouse_type ) ;
  
  vesa_show_bmp_picture( x , y , mouse_picture_addr , ( unsigned short )0xffffffff , 1 ) ;
}

/* ƶ mouse  */
void mouse_move_mouse( int x1 , int y1 , int x2 , int y2 , enum mouse_type_enum mouse_type )
{
  // Ȼָǰ mouse ڵʾǸͼ
  mouse_restore_picture( x1 , y1 ) ;
  
  // ʾ mouse
  mouse_show_mouse( x2 , y2 , mouse_type ) ;
}

/* ָλõͼƬ */
void mouse_save_picture( int x , int y )
{
  // mouse ĴС̶Ϊ 32 * 32
  vesa_copy_picture_from_screen( x , y , ( unsigned short * )mouse_old_picture , 32 , 32 ) ;
}

/* ָָλõͼƬ */
void mouse_restore_picture( int x , int y )
{
  vesa_copy_picture_to_screen( x , y , ( unsigned short * )mouse_old_picture , 32 , 32 ) ;
}
