#include "vesa.h"
#include "bmp.h"

/*****************************************************
*                                        *
*****************************************************/

const int VESA_SCREEN_X_MAX = 800 ;
const int VESA_SCREEN_Y_MAX = 600 ;

static const int VESA_ADDR = 0x20001 ;

void vesa_print_number( int number , int x , int y , unsigned short color )
{
  int pos ;
  if( number < 0 ){
    pos = '-' * 16 ;
    number = 0 - number ;
    vesa_print_english( x , y , pos , color ) ;
    x += 8 ;
  }

  if( number == 0 ){
    pos = '0' * 16 ;
    vesa_print_english( x , y , pos , color ) ;
    return ;
  }

  int tmp[ 10 ] ;
  int length = 0 ;

  while( number ){
    tmp[ length++ ] = number % 10 ;
    number = number / 10 ;
  }
  for( int i = length - 1 ; i >= 0 ; --i ){
    pos = ( tmp[ i ] + '0' ) * 16 ;
    vesa_print_english( x , y , pos , color ) ;
    x += 8 ;
  }
}

void vesa_print_string( char *string , int x , int y , unsigned short color )
{
  unsigned int pos ;
  while( *string ){
    unsigned char ch = *string++ ;
    if( ch & 0x80 ){
      // ʾ
      unsigned char hi = *string++ ;
      int q = ch - 0xa0 ;
		  int w = hi - 0xa0 ;
		  pos = ( 94 * ( q - 1 ) + ( w - 1 ) ) * 32 ;
      vesa_print_chinese( x , y , pos , color ) ;
      x += 16 ;
    }
    else{
      // ʾӢ
      pos = ch * 16 ;
      vesa_print_english( x , y , pos , color ) ;
      x += 8 ;
    }
  }
}

void vesa_print_string_in_rect( char *string , int x1 , int y1 , int x2 , int y2 , unsigned short color , int x_offset , int y_offset )
{
  unsigned int pos ;
  int x = x1 + x_offset ;
  int y = y1 + y_offset ;
  
  while( *string ){
    unsigned char ch = *string++ ;
    if( ch & 0x80 ){
      unsigned char hi = *string++ ;
      int q = ch - 0xa0 ;
		  int w = hi - 0xa0 ;
		  pos = ( 94 * ( q - 1 ) + ( w - 1 ) ) * 32 ;
      if( y + 18 > y2 ){
        return ;
      }
      if( y >= y1 && x >= x1 && x + 16 <= x2 ){ // ֹӡ
        vesa_print_chinese( x , y , pos , color ) ;
      }
      x += 16 ;      
    }
    else{
      // ʾӢ
      switch( ch ){
        case '\r' : // \r = 0xd 
          continue ; // 
          break ;
        case '\n' : // \n = 0a
          y += 18 ; // 
          x = x1 + x_offset ;
          break ;
        case '\t' :
          x += 8 * 8 ; // 8
          break ;
        default :
          pos = ch * 16 ;
          if( y + 18 > y2 ){
            return ;
          }
          if( y >= y1 && x >= x1 && x + 8 <= x2  ){            
            vesa_print_english( x , y , pos , color ) ;
          }
          x += 8 ;                    
      }
    }
  }
}

// ϶һǷһ֮
int vesa_does_point_in_rect( int x , int y , int x1 , int y1 , int x2 , int y2 )
{
  if( x >= x1 && x <= x2 ){
    if( y >= y1 && y <= y2 ){
      return 1 ;
    }
  }
  return 0 ;
}

// ɫϳɺ
unsigned short vesa_compond_rgb( unsigned char r , unsigned char g , unsigned char b )
{
  union{
    unsigned int color ;
    struct{
      unsigned char b : 5 ;
      unsigned char g : 6 ;
      unsigned char r : 5 ;
    } ;
  } u ;
  u.r = r % 32 ;
  u.g = g % 64 ;
  u.b = b % 32 ;
  return u.color ;
}

// 㺯
void vesa_draw_point( unsigned int x , unsigned int y , unsigned short color )
{  
  unsigned int x_max = VESA_SCREEN_X_MAX - 1 ; // ÿ
  unsigned int y_max = VESA_SCREEN_Y_MAX - 1 ; // ÿ

  // ֹԽ
  if( x > x_max ){
    return ;
  }
  if( y > y_max ){
    return ;
  }

  // ȡԴԵַ
  unsigned short *video = ( unsigned short * )( * ( ( unsigned int *)VESA_ADDR ) ) ;

  // ƫ
  unsigned int offset = y * ( x_max + 1 ) + x ;  
  *( video + offset ) = color ;
}

// ȡָɫ
unsigned short vesa_get_point_color( unsigned int x , unsigned int y )
{
  unsigned int x_max = VESA_SCREEN_X_MAX - 1 ; // ÿ
  unsigned int y_max = VESA_SCREEN_Y_MAX - 1 ; // ÿ

  // ֹԽ
  if( x > x_max ){
    x = x_max ;
  }
  if( y > y_max ){
    y = y_max ;
  }

  // ȡԴԵַ
  unsigned short *video = ( unsigned short * )( * ( ( unsigned int *)VESA_ADDR ) ) ;

  // ƫ
  unsigned int offset = y * ( x_max + 1 ) + x ;  
  return *( video + offset ) ;
}

// 亯
void vesa_fill_rect( unsigned int x1 , unsigned int y1 , unsigned int x2 , unsigned int y2 , unsigned int color )
{
  for( int x = x1 ; x < x2 + 1 ; ++x ){
    for( int y = y1 ; y < y2 + 1 ; ++y ){
      vesa_draw_point( x , y , color ) ;
    }
  }
}

// 
void vesa_clean_screen( unsigned short color )
{
  vesa_fill_rect( 0 , 0 , VESA_SCREEN_X_MAX - 1 , VESA_SCREEN_Y_MAX - 1 , color ) ;
}

// ͼƬ
void vesa_copy_picture_from_screen( unsigned int x , unsigned int y , unsigned short *object_picture_addr , unsigned int picture_width , unsigned int picture_height )
{
  for( int j = 0 ; j < picture_height ; ++j ){
    for( int i = 0 ; i < picture_width ; ++i ){
      *( object_picture_addr + j * picture_width + i ) = vesa_get_point_color( x + i , y + j ) ;
    }
  }
}

// ͼƬ
void vesa_copy_picture_to_screen( unsigned int x , unsigned int y , unsigned short *source_picture_addr , unsigned int picture_width , unsigned int picture_height )
{
  for( int j = 0 ; j < picture_height ; ++j ){
    for( int i = 0 ; i < picture_width ; ++i ){
      vesa_draw_point( x + i , y + j , *( source_picture_addr + j * picture_width + i ) ) ;
    }
  }
}

// ߺ
void vesa_draw_x_line( unsigned int y , unsigned int x1 , unsigned int x2 , unsigned short color )
{
  for( int x = x1 ; x < x2 + 1 ; ++x ){
    vesa_draw_point( x , y , color ) ;
  }
}

// ߺ
void vesa_draw_y_line( unsigned int x , unsigned int y1 , unsigned int y2 , unsigned short color )
{
  for( int y = y1 ; y < y2 + 1 ; ++y ){
    vesa_draw_point( x , y , color ) ;
  }
}

// κ
void vesa_draw_rect( unsigned int x1 , unsigned int y1 , unsigned int x2 , unsigned y2 , unsigned short color , int dose_fill_it )
{
  vesa_draw_x_line( y1 , x1 , x2 , color ) ;
  vesa_draw_x_line( y2 , x1 , x2 , color ) ;
  vesa_draw_y_line( x1 , y1 , y2 , color ) ;
  vesa_draw_y_line( x2 , y1 , y2 , color ) ;

  if( dose_fill_it ){
    vesa_fill_rect( x1 , y1 , x2 , y2 , color ) ;
  }
}

// ʾӢ
void vesa_print_english( unsigned int x , unsigned int y , unsigned int pos_in_font , unsigned short color )
{
  unsigned char *english_font_addr = ( unsigned char * )0x140000 + pos_in_font ; // һӢ 16 ֽ

  union{
    unsigned char ch ;
    struct{
      unsigned char ch0 : 1 ;
      unsigned char ch1 : 1 ;
      unsigned char ch2 : 1 ;
      unsigned char ch3 : 1 ;
      unsigned char ch4 : 1 ;
      unsigned char ch5 : 1 ;
      unsigned char ch6 : 1 ;
      unsigned char ch7 : 1 ;
    } ;
  } u ;

  for( int j = 0 ; j < 16 ; ++j ){ // һӢ 16     
    u.ch = english_font_addr[ j ] ;
    int offset = x ;
    if( u.ch7 ){
      vesa_draw_point( offset , y , color ) ;
    }
    if( u.ch6 ){
      vesa_draw_point( offset + 1  , y , color ) ;
    }
    if( u.ch5 ){
      vesa_draw_point( offset + 2 , y , color ) ;
    }
    if( u.ch4 ){
      vesa_draw_point( offset + 3 , y , color ) ;
    }
    if( u.ch3 ){
      vesa_draw_point( offset + 4 , y , color ) ;
    }
    if( u.ch2 ){
      vesa_draw_point( offset + 5 , y , color ) ;
    }
    if( u.ch1 ){
      vesa_draw_point( offset + 6 , y , color ) ;
    }
    if( u.ch0 ){
      vesa_draw_point( offset + 7 , y , color ) ;
    }
    y++ ; // ʾһ
  }
}

// ʾ
void vesa_print_chinese( unsigned int x , unsigned int y , unsigned int pos_in_font , unsigned short color )
{
  unsigned char *chinese_font_addr = ( unsigned char * )0x141000 + pos_in_font ; 

  union{
    unsigned char ch ;
    struct{
      unsigned char ch0 : 1 ;
      unsigned char ch1 : 1 ;
      unsigned char ch2 : 1 ;
      unsigned char ch3 : 1 ;
      unsigned char ch4 : 1 ;
      unsigned char ch5 : 1 ;
      unsigned char ch6 : 1 ;
      unsigned char ch7 : 1 ;
    } ;
  } u ;

  for( int j = 0 ; j < 16 ; ++j ){ // һ 16 
    for( int i = 0 ; i < 2 ; ++i ){ // ÿ 2 ֽ ( 16 λ )
      u.ch = chinese_font_addr[ j * 2 + i ] ;
      int offset = x + i * 8 ; // һֽ 8 
      if( u.ch7 ){
        vesa_draw_point( offset , y , color ) ;
      }
      if( u.ch6 ){
        vesa_draw_point( offset + 1  , y , color ) ;
      }
      if( u.ch5 ){
        vesa_draw_point( offset + 2 , y , color ) ;
      }
      if( u.ch4 ){
        vesa_draw_point( offset + 3 , y , color ) ;
      }
      if( u.ch3 ){
        vesa_draw_point( offset + 4 , y , color ) ;
      }
      if( u.ch2 ){
        vesa_draw_point( offset + 5 , y , color ) ;
      }
      if( u.ch1 ){
        vesa_draw_point( offset + 6 , y , color ) ;
      }
      if( u.ch0 ){
        vesa_draw_point( offset + 7 , y , color ) ;
      }    
    }
    y++ ; // ʾһ
  }
}

void vesa_read_bmp_information( void *bmp_address , int *bmp_length , int *bmp_height )
{
  struct bmp_bmp_head_struct *bmp_head = ( struct bmp_bmp_head_struct * )bmp_address ;
  *bmp_length = bmp_head->info_head.width ;
  *bmp_height = bmp_head->info_head.height ;
}

// ʾ pbmp ʽͼƬ
void vesa_show_bmp_picture( unsigned int x , unsigned int y , void *bmp_addr , unsigned short mask_color , int dose_use_mask_color )
{ 
  // ֻ֧ windows ׼ 16 λ bmp ʽͼƬ1:5:5:5
  struct bmp_bmp_head_struct *bmp_head = ( struct bmp_bmp_head_struct * )bmp_addr ;
  int width = bmp_head->info_head.width ;
  int height = bmp_head->info_head.height ;

  // 洢ÿɫʵϢڵλ
  unsigned short *color = ( unsigned short * )( ( unsigned int )bmp_addr + bmp_head->offset ) ;

  // һеֽ 4 ıˣȼÿҪ 2 Ϊÿֽ
  int fill_length = width * 2 % 4 / 2 ; 

  // bmp Ĵ˳Ǵµϣ
  for( int i = height - 1 ; i >= 0 ; --i ){
    for( int j = 0 ; j < width ; ++j ){
	  // ȡÿɫϢ
	  //  windows Ĭϵ 555 ʽ pyos õ 565 ʽȽһת 
      unsigned short temp_color = vesa_change_color_form_555_to_565( *color ) ;
      if( !dose_use_mask_color || temp_color != mask_color ){  
	    // ÿ
        vesa_draw_point( x + j , y + i , temp_color ) ;
      }
      ++color ;
    }
    // 
    color += fill_length ;
  }
}

void vesa_show_bmp_picture_in_rect( int x1 , int y1 , int x2 ,int y2 , void *bmp_addr , unsigned short mask_color , int dose_use_mask_color , int x_offset , int y_offset )
{
  // ֻ֧ windows ׼ 16 λ bmp ʽͼƬ1:5:5:5
  struct bmp_bmp_head_struct *bmp_head = ( struct bmp_bmp_head_struct * )bmp_addr ;
  int width = bmp_head->info_head.width ;
  int height = bmp_head->info_head.height ;

  // 洢ÿɫʵϢڵλ
  unsigned short *color = ( unsigned short * )( ( unsigned int )bmp_addr + bmp_head->offset ) ;

  // һеֽ 4 ıˣȼÿҪ 2 Ϊÿֽ
  int fill_length = width * 2 % 4 / 2 ; 

  // bmp Ĵ˳Ǵµϣ
  for( int i = height - 1 ; i >= 0 ; --i ){
    for( int j = 0 ; j < width ; ++j ){
	  // ȡÿɫϢ
	  //  windows Ĭϵ 555 ʽ pyos õ 565 ʽȽһת 
      unsigned short temp_color = vesa_change_color_form_555_to_565( *color ) ;
      if( !dose_use_mask_color || temp_color != mask_color ){  
	      // ÿ
        int x = x1 + j + x_offset ;
        int y = y1 + i + y_offset ;

        // ʵֲü,ֹȥ
        if( vesa_does_point_in_rect( x , y , x1 , y1 , x2 , y2 ) ){
          vesa_draw_point( x , y , temp_color ) ;
        }
      }
      ++color ;
    }
    // 
    color += fill_length ;
  }
}

// ʾַıģʽ
void vesa_print_char_under_text_model( char x )
{
  static unsigned char *video = ( unsigned char * )0xb8000 ;
  *video++ = x ;
  *video++ = 0xf ;
}

// ɫת 5:5:5 ʽתɱ׼ 5:6:5 ʽ
unsigned short vesa_change_color_form_555_to_565( unsigned short color_form_555 )
{
  //  555 мɫ * 2Ϊɫ㣬 0ĩλ 1
  union{
    unsigned int color ;
    struct{
      unsigned char b : 5 ;
      unsigned char g : 6 ;
      unsigned char r : 5 ;
    } ;
  } u1 ;
  
  union{
    unsigned int color ;
    struct{
      unsigned char b : 5 ;
      unsigned char g : 5 ;
      unsigned char r : 5 ;
      unsigned char alpha : 1 ;  // alpha ͨﲻ
    } ;
  } u2 ;
  
  u2.color = color_form_555 ;
  u1.r = u2.r ;
  if( u2.g ){
    u1.g = ( int )u2.g * 2 + 1 ;
  }
  else{
    u1.g = 0 ;
  }
  u1.b = u2.b ;
  
  return u1.color ;
}
